import sys
import math
import random


def get_args():
    # collect parameters
    if len(sys.argv) < 7:
        return print("One or more parameters in missing!")
    print(sys.argv)
    activation = sys.argv[1]
    training_algo = sys.argv[2]
    ground_file = sys.argv[3]
    distribution = sys.argv[4].strip()
    num_train = int(sys.argv[5])
    num_test = int(sys.argv[6])
    epsilon = float(sys.argv[7])

    # Testing
    # activation = "threshold"
    # training_algo = "perceptron"
    # ground_file = "ground_test.txt"
    # distribution = "bool"
    # num_train = 452
    # num_test = 300
    # epsilon = 0.2

    # Parse the Ground Truth file and generate examples
    trainingData, testData = parse_ground_file(ground_file, distribution, num_train, num_test)
    # Train the data and generate weights and theta
    weights, theta = train_data(trainingData, training_algo, activation)
    testError = test_data(testData, activation, weights, theta)
    #
    #print("weights -" + str(weights) + ",theta -" + str(theta))
    print_result(testError, num_test, epsilon)


def parse_ground_file(fileName, distribution, num_train, num_test):
    # this method will attempt to rad the ground turth file in order to generate examples
    try:
        # open file
        f = open(fileName, "r")
        lines = f.readlines()
        f.close()
        # get ground func
        ground_func = lines[0].strip()

        # generate examples for threshold function
        if ground_func.lower() == "tf":
            threshold = int(lines[1].strip())
            weights = [int(weights) for weights in (lines[2].strip().split(" "))]
            # examples generated by  generate_examples(distribution, len(weights), num_train)
            trainingData = ground_tf(generate_examples(distribution, len(weights), num_train), weights, threshold)
            testData = ground_tf(generate_examples(distribution, len(weights), num_test), weights, threshold)

        # Generate examples for disjunction function
        elif ground_func.lower() == "dis":
            positions = [int(pos) for pos in (lines[1].strip().split(" "))]
            trainingData = ground_dis(generate_examples(distribution, abs(max(positions, key=abs)), num_train),
                                      positions)
            testData = ground_dis(generate_examples(distribution, abs(max(positions, key=abs)), num_test),positions)
        # for invalid ground functions
        else:
            return print("Invalid Ground truth Function, Please check and try again!")
        return trainingData, testData
    except:
        return print("NOT PARSEABLE")


def generate_examples(distribution, n, num_examples):
    # this method generate examples for training/testing data
    try:
        exampleData = []
        if distribution.lower() != "bool" and distribution.lower() != "sphere":
            return print("Invalid Distribution!")
        else:
            exampleData = generate_vectors(num_examples, n, distribution)
            return exampleData
    except:
        return print("unable to generate examples")


def generate_vectors(num, n, dist):
    opVector = []
    try:
        while len(opVector) < num:
            ipVector = []
            for y in range(n):
                if dist.lower() == "bool":
                    r = random.randint(0, 1)
                else:
                    r = random.random()
                ipVector.append(r)
            opVector.append(ipVector)
            # normalise floating point vector to unit vector
            if dist.lower() == "sphere":
                mag = 0
                for v in ipVector:
                    mag = mag + v * v
                mag = math.sqrt(mag)
                for x in range(len(ipVector)):
                    ipVector[x] = ipVector[x] / mag
        return opVector
    except:
        return print("unable to generate vectors")


def ground_tf(trainingData, weights, threshold):
    # this method will tak ethe training data and pass the examples X through the TF in the ground file to generate y
    for j in range(len(trainingData)):
        ip = trainingData[j]
        calculatedThreshold = 0
        for i in range(len(weights)):
            calculatedThreshold += ip[i] * weights[i]
        if calculatedThreshold >= threshold:
            trainingData[j] = (trainingData[j], 1)
        else:
            trainingData[j] = (trainingData[j], 0)
    return trainingData


def ground_dis(trainingData, pos):
    # this method will tak ethe training data and pass the examples X through the DIS in the ground file to generate y
    for j in range(len(trainingData)):
        ip = trainingData[j]
        temp = 0
        for p in pos:
            sign = int(math.copysign(1, p))
            absp = abs(p) - 1
            if sign == -1:
                temp = temp or not (ip[absp])
            else:
                temp = temp or ip[absp]
        trainingData[j] = (trainingData[j], int(temp))
    return trainingData


def activation_func(activation, ip, theta):
    if activation.lower() == "relu":
        if ip < theta:
            return 0
        else:
            return 1 if ip > 0.5 else 0
    elif activation.lower() == "threshold":
        if ip < theta:
            return 0
        else:
            return 1
    elif activation.lower() == "tanh":
        return 1 if ((1 - math.exp(-2 * ip)) / (1 + math.exp(-2 * ip)))>0 else 0
    else:
        return print("Invalid Activation function")


def train_data(trainingData, tAlgo, activation):
    if tAlgo.lower() == "perceptron":
        weights = [0 for x in trainingData[0][0]]
        theta = 0
        for x in trainingData:
            ipVec = x[0]
            temp = 0
            for ip in range(len(ipVec)):
                temp = temp + ipVec[ip] * weights[ip]
            yPred = activation_func(activation, temp, theta)
            weights, theta = perceptron(yPred, x[1], weights, theta, ipVec)
        return weights, theta
    elif tAlgo.lower() == "winnow":
        weights = []
        for i in range(len(trainingData[0][0])):
            weights.append(random.randint(1, 100))
        # weights = [random.randint(1,100) for x in len(trainingData[0][0])]
        theta = 4
        for x in trainingData:
            ipVec = x[0]
            temp = 0
            for ip in range(len(ipVec)):
                temp = temp + (ipVec[ip] * weights[ip])
            yPred = activation_func(activation, temp, theta)
            weights = winnow(yPred, x[1], weights, ipVec)
        return weights, theta


def perceptron(yPred, yActual, weights, theta, ipVec):
    # implement perceptron update rule
    if yPred != yActual:
        if yPred == 1:
            for w in range(len(weights)):
                weights[w] = weights[w] - ipVec[w]
                theta += 1
        elif yPred == 0:
            for w in range(len(weights)):
                weights[w] = weights[w] + ipVec[w]
                theta -= 1
        print(",".join(map(str, ipVec)) + ":" + str(yActual) + ":update")
    else:
        print(",".join(map(str, ipVec)) + ":" + str(yActual) + ":no update")
    return weights, theta


def winnow(yPred, yActual, weights, ipVec):
    if yPred != yActual:
        alpha = 2
        if yPred > yActual:
            for w in range(len(weights)):
                weights[w] = pow(alpha, -ipVec[w]) * weights[w]
        else:
            for w in range(len(weights)):
                weights[w] = pow(alpha, ipVec[w]) * weights[w]

        print(",".join(map(str, ipVec)) + ":" + str(yActual) + ":update")
    else:
        print(",".join(map(str, ipVec)) + ":" + str(yActual) + ":no update")
    return weights


def test_data(testData, activation, weights, theta):
    testError = 0
    for x in testData:
        ipVec = x[0]
        temp = 0
        for ip in range(len(ipVec)):
            temp = temp + ipVec[ip] * weights[ip]
        yPred = activation_func(activation, temp, theta)
        testError += calculate_error(yPred, x[1], ipVec)
    return testError


def calculate_error(yPred, yActual, ipVec):
    if yPred != yActual:
        testError = abs(yPred - yActual)
    else:
        testError = 0
    print(",".join(map(str, ipVec)) + ":" + str(yPred) + ":" + str(yActual) + ":" + str(testError))
    return testError


def print_result(testError, num_test, epsilon):
    avgError = testError / num_test
    avgErrorStr = "Average Error:" + str(avgError)
    epsilonStr = "Epsilon:" + str(epsilon)
    result = ""
    if avgError < epsilon:
        result = "TRAINING SUCCEEDED"
    else:
        result = "TRAINING FAILED"
    print(avgErrorStr, "\n", epsilonStr, "\n", result)


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    get_args()
